/*
 * CDDUP.C
 *
 * CD Duplication by majority vote
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

char **name_array;
int *count_array;
char **memptr_array;
char *buf_array;
FILE **fh_array;
FILE *image_handle;
int passcount = 5;
const int sector_size = 2352;

void cleanup(void)
{
	int i;

	if (name_array) {
		for (i = 0; i < passcount; i++)
			if (name_array[i])
				free(name_array[i]);
		free(name_array);
	}

	if (fh_array) {
		for (i = 0; i < passcount; i++)
			if (fh_array[i])
				fclose(fh_array[i]);
		free(fh_array);
	}
	
	if (buf_array)
		free(buf_array);

	if (memptr_array)
		free(memptr_array);

	if (image_handle)
		fclose(image_handle);

}

/*
 * Get the last LBA of the CD
 */
int get_LBA_count(char *ID)
{
	int last_lba;
	char *tocname;
	char buf[256];
	FILE *f;

	/* Find last block number on CD */
	tocname = tempnam(".", "toc");
	sprintf(buf, "cdtools /toc /batch /id=%s >%s", ID, tocname);
	system(buf);
	f = fopen (tocname, "r");
	if (!f) {
		printf("can't open temp file %s\n", tocname);
		cleanup();
		exit(10);
	}
	while (!feof(f) && fscanf(f, "%*s (LBA %d)", &last_lba) != 2);
	fclose(f);
	unlink(tocname);
	return last_lba;
}

void main(int argc, char *argv[])
{
	int i, j, k;
	int start_block = 0, end_block, last_block, current_end_block, min_count_max;
	int append = 0;
	int blocksize = 10000;
	int memblocksize = 1000 * sector_size;
	int track = 0;
	char *ID = 0;
	char *tmpdir = 0;
	char *image_name;
	char buf[256];

	if (argc < 2) {
		printf(
			"Usage: %s\n [/N <num>] [/B <blocksize>]"
			" [/START <start_LBA>] [/END <end_LBA>] [/APPEND]\n"
			" [/M <memblocksize>] [/ID <id>] [/T <tmpdir>] <image>\n\n"
			"This tool requires GoldenHawk's DAO to be"
			" installed on the path.\n", argv[0]);
		cleanup();
		exit(10);
	}

	/* scan command line */
	for (i = 1; i < argc - 1; i++) {
		if (!_stricmp(argv[i], "/n")) {
			passcount = atoi(argv[++i]);
			continue;
		} else if (!_stricmp(argv[i], "/b")) {
			blocksize = atoi(argv[++i]);
			continue;
		} else if (!_stricmp(argv[i], "/append")) {
			append = 1;
			continue;
		} else if (!_stricmp(argv[i], "/m")) {
			memblocksize = atoi(argv[++i]) * sector_size;
			continue;
		} else if (!_stricmp(argv[i], "/id")) {
			ID = argv[++i];
			continue;
		} else if (!_stricmp(argv[i], "/t")) {
			tmpdir = argv[++i];
			continue;
		} else if (!_stricmp(argv[i], "/track")) {
			track = atoi(argv[++i]);
			continue;
		} else if (!_stricmp(argv[i], "/start")) {
			start_block = atoi(argv[++i]);
			continue;
		} else if (!_stricmp(argv[i], "/end")) {
			end_block = atoi(argv[++i]);
			continue;
		} else if (!_stricmp(argv[i], "")) {
		}
	}

	if (i == argc) {
		printf("No image filename given!\n");
		cleanup();
		exit(10);
	} else {
		image_name = argv[i];
	}

	if (ID == 0) {
		printf("No CDROM ID selected!\n");
		cleanup();
		exit(10);
	}

	last_block = get_LBA_count(ID) - 1;
	if (end_block > last_block && (track == 0)) {
		end_block = last_block;
		printf("End block number truncated to %d\n", end_block);
	}

	printf("Last LBA: %d\n", last_block);

	name_array = calloc(passcount, sizeof (void *));
	if (!name_array) {
		printf("Can't allocate name_array!\n");
		cleanup();
		exit(10);
	}

	count_array = calloc(passcount, sizeof (int));
	if (!count_array) {
		printf("Can't allocate count_array!\n");
		cleanup();
		exit(10);
	}

	memptr_array = calloc(passcount, sizeof (void *));
	if (!memptr_array) {
		printf("Can't allocate memptr_array!\n");
		cleanup();
		exit(10);
	}

	fh_array = calloc(passcount, sizeof (void *));
	if (!fh_array) {
		printf("Can't allocate fh_array!\n");
		cleanup();
		exit(10);
	}

	buf_array = malloc(memblocksize * (passcount + 1));
	if (!buf_array) {
		printf("Can't allocate temporary buffers!\n");
		cleanup();
		exit(10);
	}

	/* Allocate temporary names */
	for (j = 0; j < passcount; j++) {
		char *imgname;
		imgname = tmpnam("img");
		name_array[j] = malloc(strlen(imgname) + 1);
		if (!name_array[j]) {
			printf("Can't allocate space for filename!\n");
			cleanup();
			exit(10);
		}
		strcpy(name_array[j], imgname);
	}

	/* do actual CD access */
	if (append)
		image_handle = fopen(image_name, "ab");
	else
		image_handle = fopen(image_name, "wb");
	if (!image_handle) {
		printf("Can't open image file!\n");
		cleanup();
		exit(10);
	}

	if (track)
		start_block = 0;
	for (; start_block <= end_block; start_block += blocksize) {

		if (track) {
			/* make sure we quit next loop */
			start_block = end_block + 1;
		} else {
			/* compute current block size */
			if (start_block + blocksize - 1 < end_block)
				current_end_block = start_block + blocksize - 1;
			else
				current_end_block = end_block;
		}
		/* first, read all sections to temp files */
		for (j = 0; j < passcount; j++) {
			if (track)
				sprintf(buf, "cdclip %s /batch /bin /ID=%s /TRACK=%d\n",
					name_array[j], ID, track);
			else
				sprintf(buf, "cdclip %s /batch /bin /ID=%s /START=%d /END=%d\n",
					name_array[j], ID, start_block, current_end_block);
			puts(buf);
			system(buf);
		}

		/* now open file handles for all temporary files */
		for (j = 0; j < passcount; j++) {
			fh_array[j] = fopen(name_array[j], "rb");
			if (!fh_array[j]) {
				printf("Can't open image file!\n");
				cleanup();
				exit(10);
			}
		}

		/* do actual majority vote algorithm */
		min_count_max = passcount;
		while(1) {
			int size, count_max, loc_max;

			for (j = 0; j < passcount; j++) {
				memptr_array [j] = buf_array + memblocksize * (j + 1);
				size = fread(memptr_array[j], 1, memblocksize, fh_array[j]);
			}
			if (size == 0) break;
			for (i = 0; i < size; i++) {
				for (j = 0; j < passcount; j++)
					count_array[j] = 1;
				for (j = 0; j < passcount; j++) {
					for (k = j + 1; k < passcount; k++) {
						if (memptr_array[j][i] == memptr_array[k][i])
							count_array[j]++;
					}
					// have we achived the maximum possible?
					if (count_array[j] == passcount - j)
						break;
				}
				count_max = 0;
				for (j = 0; j < passcount; j++) {
					if (count_array[j] > count_max) {
						count_max = count_array[j];
						loc_max = j;
					}
				}
				buf_array[i] = memptr_array[loc_max][i];
				if (min_count_max > count_max)
					min_count_max = count_max;
			}
			size = fwrite(buf_array, 1, size, image_handle);
		}

		/* close file handles and disable handles */
		for (j = 0; j < passcount; j++) {
			fclose(fh_array[j]);
			fh_array[j] = (void *)0;
			unlink(name_array[j]);
		}
	}

	printf("min_count_max: %d\n", min_count_max);
	cleanup();
	exit(0);
}

